unit basShell;

{ $DEFINE STDIO}

interface
uses Windows,Chip,Classes;

 procedure LoadOSEnvironment(Environment:TStrings);

type
 TExecutor=class
 private
  Line:string;
  Col:integer;
 protected
  procedure NewLine(Process: THandle; Buffer:PChar; Count:integer);virtual;
  function ReadChar(Process: THandle; var Buffer:Char):boolean;virtual;
 public
  Script:TBasic;
  Output:string;
  Capture:boolean;
  function Execute(const cmd:string):integer;
 end;

type
 TBasic=class(Chip.TBasic)
 protected
  procedure CallShellSub(const cmd:string);override;
  function CallShellFunction(const cmd:string):string;override;
  function Shell(const cmd:string):integer;
  function ShellFunction(const cmd:string):string;
 public
  constructor Create;override;
 end;

implementation
uses SysUtils,ConsoleApp;

 function TExecutor.ReadChar;
 begin
  Result:=false
 end;

 procedure TExecutor.NewLine;

  procedure TTY(ch:char);
  begin
   case ch of
   #8: if Col>1 then
        if Col>Length(Line) then begin
         dec(Col);
         SetLength(Line,Col-1)
        end else
         { TODO: };
   #10: begin if Output<>'' then Output:=Output+#13#10; Output:=Output+Line; Line:=''; Col:=1 end;
   #13: Col:=1;
   else
    if Col<=Length(Line) then
     Line[Col]:=ch
    else
     Line:=Line+ch;
    inc(Col)
   end
  end;

 begin
  if not Assigned(Script) then exit;
  if Count>0 then OemToCharBuff(Buffer,Buffer,Count);
  while Count>0 do begin
   if Capture then
    TTY(Buffer[0])
   else begin
    {$IFDEF STDIO}
    Write(output,Buffer[0]);
    {$ELSE}
    Write(Script.output,Buffer[0]);
    {$ENDIF}
   end;
   inc(Buffer);
   dec(Count)
  end
 end;

 function TExecutor.Execute;
 var h,t,n:integer;
 var qch:char;

  function GetEnv:string;
  var env:TStrings;

   procedure QuickSort(L, R: Integer);
   var
     I, J: Integer;
     P: string;
   begin
     repeat
       I := L;
       J := R;
       P := env.Names[(L + R) shr 1];
       repeat
         while CompareText(env.Names[I], P) < 0 do Inc(I);
         while CompareText(env.Names[J], P) > 0 do Dec(J);
         if I <= J then
         begin
           env.Exchange(I, J);
           Inc(I);
           Dec(J);
         end;
       until I > J;
       if L < J then QuickSort(L, J);
       L := I;
     until I >= R;
   end;

  var i:integer;
  begin
   env:=TStringList.Create;
   if Assigned(Script) then begin
    env.Assign(Script.Environment);
    Script.ExportCommons(env)
   end;

   { now sort the env block since OS relies on it }
   if env.Count>1 then QuickSort(0,env.Count-1);

   Result:='';
   for i:=0 to env.Count-1 do
    Result:=Result+env[i]+#0;

   env.Free
  end;

 begin
  { separate command from parameters }
  n:=length(cmd);
  h:=1;
  while (h<=n) and (cmd[h]=' ') do inc(h);
  if (h<=n) and (cmd[h]='"') then begin
   inc(h);
   qch:='"'
  end else
   qch:=' ';
  t:=h;
  while (t<=n) and (cmd[t]<>qch) do inc(t);

  Line:='';
  Output:='';
  Col:=1;
  Result:=ExecConsoleApp(Copy(cmd,h,t-h),Copy(cmd,t+1,maxint),GetEnv,ReadChar,NewLine);
  if Line<>'' then begin
   if Output<>'' then Output:=Output+#13#10;
   Output:=Output+Line;
   Line:=''
  end
 end;

 procedure LoadOSEnvironment;
 var env,p:PChar;
 var n:integer;
 var line:string;
 begin
  Environment.Clear;
  env:=GetEnvironmentStrings;
  p:=env;
  while p[0]<>#0 do begin
   n:=StrLen(p);
   SetLength(line,n);
   Move(p[0],line[1],n);
   Environment.Add(line);
   inc(p,n+1)
  end;
  FreeEnvironmentStrings(env)
 end;

 function TBasic.Shell;
 var exec:TExecutor;
 begin
  exec:=TExecutor.Create;
  exec.Script:=Self;
  try
   Result:=exec.Execute(cmd)
  finally
   exec.Free
  end
 end;

 function TBasic.ShellFunction;
 var exec:TExecutor;
 begin
  exec:=TExecutor.Create;
  exec.Script:=Self;
  exec.Capture:=true;
  try
   exec.Execute(cmd);
   Result:=exec.Output
  finally
   exec.Free
  end
 end;

 procedure TBasic.CallShellSub;
 begin
  Shell(cmd)
 end;

 function TBasic.CallShellFunction;
 begin
  Result:=ShellFunction(cmd)
 end;

 constructor TBasic.Create;
 begin
  inherited;
  LoadOSEnvironment(Environment);
  MapMethod('Shell',vkInteger,chr(vkString or vkConst),Self,@TBasic.Shell);
  MapMethod('Shell$',vkString,chr(vkString or vkConst),Self,@TBasic.ShellFunction);
 end;

end.
